#!/usr/bin/env python
#input: m,n,p,size of matrix A, size of matrix B
#output: the times for different stage

from mpi4py import MPI
import math
import numpy as np
import time
import sys
#import mpmath as mp

############################################
########### Master's function   ############
############################################

def encodeData(A, B, P, Delta, worldSize):
    Ap = []
    Bp = []
    shapeA = A.shape
    shapeB = B.shape
    for i in range(1, worldSize):
        f = np.zeros((shapeA[0] // P[0], shapeA[1] // P[1]))
        g = np.zeros((shapeB[0] // P[1], shapeB[1] // P[2]))
        for x0 in range(P[0]):
            for x1 in range(P[1]):
                f = f +  A[(x0 * shapeA[0] // P[0]): ((x0 + 1) * shapeA[0] // P[0]), (x1 * shapeA[1] // P[1]): ((x1 + 1) * shapeA[1] // P[1])] * (Delta[i-1] ** (P[1] * P[2] * x0 + x1))
        Ap.append(f)
        for x2 in range(P[2]):
            for x1 in range(P[1]):
                g = g +  B[((P[1] - 1 -  x1) * shapeB[0] // P[1]): ((P[1] - 1 - x1 + 1) * shapeB[0] // P[1]), (x2 * shapeB[1] // P[2]): ((x2 + 1) * shapeB[1] // P[2])] * (Delta[i-1] ** (P[1] * x2 + x1))
        Bp.append(g)
    return Ap, Bp

def sendCodedTask(Ap, Bp, comm):
    for i in range(1, comm.Get_size()):
        comm.send(Ap[i - 1], dest=i, tag=(100+i))
        comm.send(Bp[i - 1], dest=i, tag=(200+i))

def recvResult(Delta, threshold, comm):
    indexData = []
    result = []
    R = []
    newDelta = []
    NewDelta = []
    for i in range(comm.Get_size()):
        R.insert(i, "")
        NewDelta.insert(i, "")

    while(len(indexData) < threshold):
        st = MPI.Status()
        comm.Probe(status = st)
        i = st.source
        R.insert(i - 1, comm.recv(source=i, tag=2))
        del R[i]
        receiveNewDelta = Delta[i-1]
        if P_new[1] != P[1]:
            receiveNewDelta = pow(receiveNewDelta, P[1] / P_new[1])
        NewDelta.insert(i-1, receiveNewDelta)
        del NewDelta[i]
        indexData.append(i)
    indexData.sort()
    for j in indexData:
        result.append(R[j-1])
        newDelta.append(Delta[j-1])
    return result, newDelta

def traditional_decode(P_old, R, P, X, shapeA):
    X = np.array(X, dtype=np.float)
    m_old = P_old[0]
    n_old = P_old[1]
    m = P[0]
    n = P[2]
    lambda_m = m // m_old
    lambda_n = n // n_old
    threshold = P[0] * P[1] * P[2] + P[1] - 1
    M = np.ones((threshold, threshold))
    power = np.prod(P[1 : 2 + 1]) + np.prod(P[1 : 2])
    base = power
    power += np.prod(P[2 : 2]) * 2
    base += np.prod(P[2 : 2]) * (P[1] + 1)
    base = base - power
    for i in range(0, threshold):
        for j in range(0, threshold):
            M[i, j] = float(X[i]) ** j
    rows = []
    for x0 in range(P[0]):
        for xm in range(P[2]):
            rows.append(base + np.prod(P[1 : 2 + 1]) * x0 + np.prod(P[1 : 2]) * xm)
    rows = [int(x) for x in rows]
    R = np.reshape(R, (threshold, -1))
    M = np.linalg.pinv(M)
    M = M[rows, :]
    result1 = np.dot(M,R)
    Result =[]
    for i in range(m * n):
        f = np.reshape(result1[i],(shapeA[0]//m,-1))
        Result.append(f)
    if lambda_m != 1:
        R = []
        for j in range(m*n//lambda_m):
            for i in range(lambda_m):
                R.append(Result[m*n//lambda_m*i+j])
        R = np.array(R)
        Result = R
    if lambda_n != 1:
        Re = []
        for i in range(m):
            for j in range(n_old):
                for k in range(lambda_n):
                    Re.append(Result[n*i+j+n_old*k])
        Re = np.array(Re)
        Result = Re
    Result = np.vstack(np.split(np.hstack(Result), m, 1))
    return Result                

def print_time(start):
    end = time.time()
    print (str(end - start) + ' seconds')

#we use this function to avoid stragglers being suspend in each repeat, so we can use shell scripts to run this code for 50 times.
def discStragglers(threshold, comm):
    world = comm.Get_size()
    stragglers = world - 1 - threshold

    for i in range(stragglers):
        st = MPI.Status()
        comm.Probe(status = st)
        i = st.source
        data = comm.recv(source=i, tag=2)

def computeTime(start):
    current = time.time()
    t = current - start
    return t        

###########################################
##########  worker's function   ###########
###########################################

def EP_re_encode_p(M, P, lambda_p, t):
    S = []
    for i in range(2):
        S.append(M[i].shape)
    g = []
    g.append(np.zeros((S[0][0], S[0][1] // lambda_p)))
    g.append(np.zeros((S[1][0] // lambda_p, S[1][1])))
    #n = pow(t, 1.0/lambda_p)
    n = np.power(t, 1.0/lambda_p)
    A = np.hsplit(M[0], lambda_p)
    B = np.vsplit(M[1], lambda_p)
    for i in range(lambda_p):
        g[0] = g[0] + A[i] * (n ** i)
        g[1] = g[1] + B[(lambda_p - 1 - i)] * (n ** i)
    t = n
    return g, n

def EP_re_encode_m(M, P, lambda_m, t):
    S = []
    for i in range(2):
        S.append(M[i].shape)
    g = []
    g.append(np.zeros((S[0][0] // lambda_m, S[0][1])))
    g.append(np.zeros((S[1][0], S[1][1])))
    C = np.vsplit(M[0], lambda_m)
    for i in range(lambda_m):
        g[0] = g[0] + C[i] * (t ** (i * P[0] * P[1] * P[2]))
    g[1] = M[1]
    return g

def EP_re_encode_n(M, P, lambda_n, t):
    S = []
    for i in range(2):
        S.append(M[i].shape)
    g = []
    g.append(np.zeros((S[0][0], S[0][1])))
    g.append(np.zeros((S[1][0], S[1][1] // lambda_n)))
    D = np.hsplit(M[1], lambda_n)
    for i in range(lambda_n):
        g[1] = g[1] + D[i] * (t ** (i * P[0] * P[1] * P[2]))
    g[0] = M[0]
    return g

def EP_re_encode(M, P_new, P, t):
    if (P[0] != P_new[0]):
        M = EP_re_encode_m(M, P, P_new[0] // P[0], t)
    if (P[2] != P_new[2]):
        M = EP_re_encode_n(M, [P_new[0], P[1], P[2]], P_new[2] // P[2], t)
    if (P[1] != P_new[1]):
        M, t = EP_re_encode_p(M, [P_new[0], P[1], P_new[2]], P_new[1] // P[1], t)
    return M, t

def recvCodedTask(comm):
    rank = comm.Get_rank()
    Ap = comm.recv(source=0, tag=100 + rank)
    Bp = comm.recv(source=0, tag=200 + rank)
    return Ap, Bp

def matmul(Ap, Bp):
    Cp = np.matmul(Ap, Bp)
    return Cp

def uploadResult(Cp, comm):
    comm.send(Cp, dest=0, tag=2)

###########################################
############# main function   #############
###########################################

#Parameters of MPI
comm = MPI.COMM_WORLD
worldSize = comm.Get_size()
rank = comm.Get_rank()
processorName = MPI.Get_processor_name()
TaskMaster = 0
#Input
m = int(sys.argv[1])
n = int(sys.argv[2])
p = int(sys.argv[3])
m_new = int(sys.argv[4])
n_new = int(sys.argv[5])
p_new = int(sys.argv[6])
P_new = [m_new, p_new, n_new]
P = [m, p, n]
size_1 = int(sys.argv[7])
size_2 = int(sys.argv[8])
size_3 = int(sys.argv[9])
Delta = np.linspace(-1.0, 1.0, num= (worldSize - 1))

if rank == TaskMaster:
    start_total = time.time()
    #Different methods for loading data
    A = np.random.rand(size_1, size_2)
    B = np.random.rand(size_2, size_3)
    #A = np.load('./job1/A.npy')
    #B = np.load('./job1/B.npy')
    #A = np.ones((size_1, size_2))
    #B = np.ones((size_2, size_3))
    #A = np.matrix(np.random.randint(0,10, (size_1, size_2)))
    #B = np.matrix(np.random.randint(0,10, (size_2, size_3)))
    time_loading = computeTime(start_total)
    shapeA = A.shape
    threshold = P_new[0] * P_new[1] * P_new[2] + P_new[1] -1
    start_encode = time.time()
    Ap, Bp = encodeData(A, B, P, Delta, worldSize)
    time_encoding = computeTime(start_encode)
    start_send = time.time()
    sendCodedTask(Ap, Bp, comm)
    #for computing the time of sending
    comm.barrier() 
    time_sending = computeTime(start_send)
    start_reencode = time.time()
    #for computing the time of re-encoding
    comm.barrier() 
    time_reencoding = computeTime(start_reencode)    
    start_recv = time.time()
    result, newDelta = recvResult(Delta, threshold, comm)
    time_recving = computeTime(start_recv)
    start_recvstrag = time.time()
    discStragglers(threshold, comm)
    time_recvstrag = computeTime(start_recvstrag)
    start_decode = time.time()    
    Result = traditional_decode(P, result, P_new, newDelta, shapeA)
    time_decoding = computeTime(start_decode)
    time_total = computeTime(start_total)    
    print(time_loading)
    print(time_encoding)
    print(time_sending)
    print(time_reencoding)
    print(time_recving)
    print(time_recvstrag)
    print(time_decoding)
    print(time_total)
    #To compute the data we demonstrated in paper
    jobCompletionTime = time_reencoding + time_recving + time_decoding
    overhead = time_reencoding
    print(jobCompletionTime)
    print(overhead)
    exit(0)

if rank != TaskMaster:
    Ap, Bp = recvCodedTask(comm)
    f = [Ap, Bp]
    #for computing the time of sending
    comm.barrier()
    delta = Delta[rank-1]
    F = EP_re_encode(f, P_new, P, delta)
    Cp = matmul(F[0][0], F[0][1])
    #for computing the time of re-encoding
    comm.barrier()
    uploadResult(Cp, comm)    
